/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifndef WIN32
	#define _cdecl
	#undef PERF
#else
//	#define PERF
#endif

#ifdef PERF
#include <windows.h>
typedef __int64 int64;

class cHiresStopWatch
{
private:
	int64 iStart,iEnd;
	
	static int64 freq;

public:
	static bool Init(void)
	{
		return (0==::QueryPerformanceFrequency((LARGE_INTEGER*)&cHiresStopWatch::freq));
	}

public:
	cHiresStopWatch()
		:iStart(0)
		,iEnd(0)
	{
		this->Start();
	}

	void Start()
	{
		::QueryPerformanceCounter((LARGE_INTEGER*)&this->iStart);
	}

	void Stop()
	{
		::QueryPerformanceCounter((LARGE_INTEGER*)&this->iEnd);
	}

	int64 Freq() const
	{
		return freq;
	}

	int64 DeltaTicks() const{return (this->iEnd-this->iStart);}
	double DeltaSec() const {return (double)this->DeltaTicks()/(double)this->Freq();}
	double DeltaMilli() const {return 1000.0*(double)this->DeltaTicks()/(double)this->Freq();}
	double DeltaMicro() const {return 1000000.0*(double)this->DeltaTicks()/(double)this->Freq();}
};
int64  cHiresStopWatch::freq;
#endif

const char *dictname="web2.dict";

char *rawdict=0;

bool matchsubstr(const char *szP, const char *szT)
{
	int i=0;
	char cp;
	while ((cp=szP[i]) & szT[i])
	{
		if(cp != '?')
		{
			if (cp != szT[i])
				return false;
		}

		++i;
	}

	return !cp && !szT[i];
}

class Dictionary
{
private:
	void _alloc(long wordcount)
	{
		this->words = (const char **) malloc(sizeof(const char*) * wordcount);
		if (this->words)
			this->count = wordcount;
	}

	void _zerohistogram(void)
	{
		for (int i=0; i<256; ++i)
			this->hist[i] = 0;
	}

	void _realloc(void)
	{
		this->words = (const char**) realloc(this->words, sizeof(const char*) * this->count);
	}

	void _copy(const Dictionary& d)
	{
		this->_alloc(d.count);
		if (this->words)
		{
			::memcpy(this->words, d.words, this->count* sizeof(const char*));
			::memcpy(this->hist, d.hist, sizeof(this->hist));
		}
	}

public:
	long count;
	const char **words;
	long hist[256];

	Dictionary()
		:words(0)
		,count(0)
	{
		this->_zerohistogram();
	}

	Dictionary(const Dictionary& d)
	{
		this->_copy(d);
	}

	void operator=(const Dictionary& d)
	{
		this->_copy(d);
	}

	~Dictionary()
	{
		if (this->words)
			free(this->words);	
	}

	void LoadRaw(const char* szRaw, long wordcount)
	{
		this->words = (const char **) malloc(sizeof(const char*) * wordcount);
		if (this->words)
		{
			this->count = wordcount;

			const char *sz = szRaw;
			const char **rgw = this->words;
			 
			for (long i=0; i<wordcount; ++i)
			{
				*rgw++ = sz;
				sz += ::strlen(sz)+1;
			}
		}
	}
};

Dictionary dcFull;


bool filelen(const char* fn, size_t * plen)
{
	struct stat buf;
	if (0 == stat(fn, &buf))
	{
		*plen = buf.st_size;
		return false;
	}

	return true;
}

bool loaddict(void)
{
	bool fErr;
	size_t len;
	
	fErr = filelen(::dictname, &len);
	if(!fErr)
	{
		::rawdict = (char*)malloc(len+1);
		if (::rawdict)
		{
			FILE * pf;
			
			pf = fopen(::dictname, "r");
			if (pf)
			{
				size_t l;
			
				l = fread(::rawdict, 1, len, pf);

				//fErr = (len != l);
				len = l;
				fclose(pf);
			}

			if (!fErr)
			{
				size_t i;
				
				long wordcount=0;
				for (i=0; i<len; ++i)
				{
					if (::rawdict[i] == '\n')
					{
						::rawdict[i] = 0;
						++wordcount;
					}
				}
				::rawdict[i] = 0;
				::dcFull.LoadRaw(::rawdict, wordcount);
			}
		}
	}


	return fErr;
}


#ifdef PERF
cHiresStopWatch clk;
#endif

bool matchword(const char* szSeed, const char *szCandidate)
{
	long nSeed = ::strlen(szSeed);
	long nCand = ::strlen(szCandidate);
	if (nCand<4 || nCand > nSeed || ::isupper(*szCandidate))
		return false;

	int i;
	char *szC = (char*)::alloca(nCand+1);
	for(i=0; i<nCand && szCandidate[i]; ++i)
		szC[i] = ::tolower(szCandidate[i]);
	szC[i]=0;
	char *szS = (char*)::alloca(nSeed+1);
	for(i=0; i<nSeed && szSeed[i]; ++i)
		szS[i] = ::tolower(szSeed[i]);
	szS[i]=0;

	if (!::strchr(szC, *szS))
		return false;

	while(*szC)
	{
		char c = ::tolower(*szC++);
		for(char *s = szS; *s; ++s)
		{
			if (c == *s)
			{
				while(*s)
				{
					*s = *(s+1);
					++s;
				}
				goto continuescan;
			}
		}
		return false;

continuescan:
		if (*szS)
			continue;
		else
			return true;
	}

	return true;
}

void _cdecl main(int argc, char *argv[])
{
	char *szword="";

	if (argc == 2)
		szword = argv[1];
	else
		fprintf(stderr, "No letters to match\n");

	if (loaddict())
		fprintf(stderr, "Error opening dictionary file\n");
	else
	{
#ifdef PERF
		cHiresStopWatch::Init();
#endif
	
		if (*szword)
		{
			for (long i=0;i<::dcFull.count; ++i)
			{
				if (matchword(szword, ::dcFull.words[i]))
				{
					printf("%s\n",  ::dcFull.words[i]);
				}
			}
		}
		else
		{
#ifdef PERF
#endif
		}
	}
}
